/*
 * Copyright (c) 1997
 *      Sandia National Laboratories.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory, Sandia National Laboratories, and its
 * contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <netdissect-stdinc.h>

#include "netdissect.h"
#include "addrtoname.h"
#include "extract.h"

#if SIZEOF_LONG_LONG == 8
/*
 * XTP sequence numbers are 64 bit integers operated
 * on with modular arithmetic.  These macros can be
 * used to compare such integers.
 */
#define XTP_SEQ_LT(a,b)     ((uint64_t)((a)-(b)) < (uint64_t)0)
#define XTP_SEQ_LEQ(a,b)    ((uint64_t)((a)-(b)) <= (uint64_t)0)
#define XTP_SEQ_GT(a,b)     ((uint64_t)((a)-(b)) > (uint64_t)0)
#define XTP_SEQ_GEQ(a,b)    ((uint64_t)((a)-(b)) >= (uint64_t)0)
#define XTP_SEQ_MAX(a,b)    (XTP_SEQ_LT(a,b)?(b):(a))
#define XTP_SEQ_MIN(a,b)    (XTP_SEQ_LT(a,b)?(a):(b))

/*
 * XTP keys are 64 bit integers operated
 * on with modular arithmetic.  These macros can be
 * used to compare such integers.
 */
#define XTP_KEY_LT(a,b)     ((uint64_t)((a)-(b)) < (uint64_t)0)
#define XTP_KEY_LEQ(a,b)    ((uint64_t)((a)-(b)) <= (uint64_t)0)
#define XTP_KEY_GT(a,b)     ((uint64_t)((a)-(b)) > (uint64_t)0)
#define XTP_KEY_GEQ(a,b)    ((uint64_t)((a)-(b)) >= (uint64_t)0)

/*
 * Most other XTP fields are 32 bit integers operated
 * on with modular arithmetic.  These macros can be
 * used to compare such integers.
 */
#define XTP_LT(a,b)     ((int)((a)-(b)) < 0)
#define XTP_LEQ(a,b)    ((int)((a)-(b)) <= 0)
#define XTP_GT(a,b)     ((int)((a)-(b)) > 0)
#define XTP_GEQ(a,b)    ((int)((a)-(b)) >= 0)
#define XTP_MAX(a,b)    (XTP_LT(a,b)?(b):(a))
#define XTP_MIN(a,b)    (XTP_LT(a,b)?(a):(b))

#define keytoindex(key) ((int)(((u_long)(key)) & (MAX_CONTEXTS - 1)))

/*
 * Macros to set/clr/check 64th bit
 */
#define TEST_HI_BIT(d) ((int)((uint64_t)(d) >> 63))
#define SET_HI_BIT(d)  ((uint64_t)(d) | (uint64_t)1 << 63)
#define CLR_HI_BIT(d)  (((uint64_t)(d) << 1) >> 1)

typedef uint64_t  xtp_key;
typedef uint64_t  xtp_seq;

/*
 * The XTP command field has two parts, options (3 bytes) and the packet
 * type (1 byte). The structure below takes advantage of the fact that
 * the lower byte of the options subfield is not currently used.
 */
typedef struct {
    u_short options;        /* XTP options bit flags */
    u_char  resv;           /* Reserved flag field */
    u_char  ptype;          /* Packet type */
} xtp_cmd;

/*
 * XTP options bits
 */
#define XH_BTAG     0x0001      /* btag field present in data */
#define XH_END      0x0002      /* end of association */
#define XH_EOM      0x0004      /* end of message */
#define XH_WCLOSE   0x0008      /* writer process closed */
#define XH_RCLOSE   0x0010      /* reader process closed */
#define XH_DREQ     0x0020      /* request status upon data delivery */
#define XH_SREQ     0x0040      /* request for status immediately */
#define XH_FASTNAK  0x0080      /* fast negative acknowledgement */
#define XH_NOFLOW   0x0100      /* disable flow control */
#define XH_SORT     0x0200      /* sort value valid */
#define XH_RES      0x0400      /* reservation mode */
#define XH_MULTI    0x0800      /* multicast mode */
#define XH_NOERR    0x1000      /* error recovery disabled */
#define XH_EDGE     0x2000      /* send cntl on edge trigger */
#define XH_NOCHECK  0x4000      /* disable checksum over data */

/*
 * Version number for compatibility checking
 */
#define X_VER   0x20        /* XTP 4.0 version number */

/*
 * The ptype field of the xtp_cmd structure is a byte with the top
 * 3 bits are the version number (0x20 is value "1") and the lower
 * 5 bits are the packet type. Here we concatonate the version with
 * the packet type value
*/
#define X_DATA  0x20        /* DATA packet (0) */
#define X_CNTL  0x21        /* CNTL packet (1) */
#define X_FIRST 0x22        /* FIRST packet (2) */
#define X_ECNTL 0x23        /* ECNTL packet (3) */
#define X_TCNTL 0x25        /* TCNTL packet (5) */
#define X_JCNTL 0x27        /* JCNTL packet (7) */
#define X_DIAG  0x28        /* DIAG packet (8) */

/*
 * XTP header.
 * Per XTP 95-21, March, 1995.
 */
struct xtphdr {
    xtp_key xh_key;         /* key */
    xtp_cmd xh_cmd;         /* command field */
    u_long  xh_dlen;        /* data length */
    u_short xh_check;       /* checksum */
    u_short xh_sort;        /* sort */
    u_long  xh_sync;        /* sync */
    xtp_seq xh_seq;         /* sequence number */
};

/*
 * Header length is a constant, so define here for efficiency
 */
#define X_HDR_LEN   32  /* 32 bytes */

/*
 * XTP address segment for IP
 */
struct xtpaddrseg {
    u_short xa_len;     /* address segment length (2) */
    u_char xa_domain;   /* address domain (1) */
    u_char xa_format;   /* address segment format (1) */
    struct in_addr xa_dst;  /* destination host address (4) */
    struct in_addr xa_src;  /* source host address (4) */
    u_short xa_dport;   /* destination port (2) */
    u_short xa_sport;   /* source port (2) */
};

/*
 * Service types
 */
#define XA_UNSPEC   0x00    /* Unspecified */
#define XA_UNACKDG  0x01    /* Unacknowledged Datagram Service */
#define XA_ACKDG    0x02    /* Acknowledged Datagram Service */
#define XA_TRANS    0x03    /* Transaction Service */
#define XA_RELSTRM  0x04    /* Reliable Unicast Stream Service */
#define XA_UNACKMC  0x05    /* Unacknowledged Multicast Stream Service */
#define XA_RELMC    0x06    /* Reliable Multicast Stream Service */

/*
 * Traffic Specifier 0: the null specifier
 */
struct xtptraf0 {
    u_long pad;     /* pad to meet 8-byte boundary */
};

/*
 * Traffic Specifier 1: the normal XTP specifier
 */
struct xtptraf1 {
    u_long maxdata;     /* maximum Information Segment size (4) */
    u_long inrate;      /* incoming rate (4) */
    u_long inburst;     /* incoming burst (4) */
    u_long outrate;     /* outgoing rate (4) */
    u_long outburst;    /* outgoing burst (4) */
};

/*
 * XTP Traffic Specifier
 */
struct xtptrafspec {
    u_short xt_len;     /* traffic segment length (2) */
    u_char xt_service;  /* type of service (1) */
    u_char xt_format;   /* traffic segment format (1) */
    union {
        struct xtptraf0 xt_ts0;
        struct xtptraf1 xt_ts1;
    } xt_tspec;
};

/*
 * XTP Traffic Segment
 */
struct xtptrafseg {
    u_long xt_rsvd;     /* padding for alignment */
    xtp_key xt_xkey;    /* exchange key */
    u_short xt_len;     /* traffic segment length (2) */
    u_char xt_service;  /* type of service (1) */
    u_char xt_format;   /* traffic segment format (1) */
    union {
        struct xtptraf0 xt_ts0;
        struct xtptraf1 xt_ts1;
    } xt_tspec;
#define xt_ts0 xt_tspec.xt_ts0
#define xt_ts1 xt_tspec.xt_ts1
};

/*
 * XTP Control Segment Fields
 */

/*
 * Spans
 */
struct xtpspan {
    xtp_seq span_a;
    xtp_seq span_b;
};

/*
 * Normal control segment
 */
struct xtpcntl {
    xtp_seq xc_rseq;    /* received sequence number (8) */
    xtp_seq xc_alloc;   /* allocation (upper edge of window) (8) */
    u_long xc_echo;     /* for echoing the sync value (4) */
};

/*
 * Error-reporting control segment
 */
struct xtpecntl {
    xtp_seq xc_rseq;    /*  ^                  */
    xtp_seq xc_alloc;   /*  |  same as common  */
    u_long xc_echo;     /*  v                  */
    u_long xc_nspan;    /* number of spans (4) */
    struct xtpspan xc_spans;/* spans (nspans * 16) */
};

/*
 * Traffic-shaping control segment
 */
struct xtptcntl {
    xtp_seq xc_rseq;    /*  ^                  */
    xtp_seq xc_alloc;   /*  |  same as common  */
    u_long xc_echo;     /*  v                  */
    u_long xc_rsvd;     /* alignment */
    xtp_key xc_xkey;    /* exchange key field (8) */
};

/*
 * Multicast join control segment
 */
struct xtpjcntl {
    xtp_seq xc_rseq;    /*  ^                  */
    xtp_seq xc_alloc;   /*  |  same as common  */
    u_long xc_echo;     /*  v                  */
    u_long xc_rsvd;     /* alignment */
    xtp_key xc_xkey;    /* exchange key field (8) */
};

/*
 *  Management Segment Fields
 */
struct xtpmgnt {
    u_long xm_code;     /* code for diag */
    u_long xm_val;      /* val for diag */
    char xm_msg[128];   /* text of the diag message */
};

struct xtphdr_long {
    struct xtphdr xp_header;
#define xh_key xp_header.xh_key
#define xh_cmd xp_header.xh_cmd
#define xh_dlen xp_header.xh_dlen
#define xh_check xp_header.xh_check
#define xh_sort xp_header.xh_sort
#define xh_sync xp_header.xh_sync
#define xh_seq xp_header.xh_seq
        union {
        struct xtpaddrseg addr; /* address segment */
        struct xtpcntl cntl;    /* common control segment */
        struct xtptcntl tcntl;  /* Traffic-shaping segment */
    } xp_seg;
#define xp_addr xp_seg.addr
#define xp_cntl xp_seg.cntl
#define xp_tcntl xp_seg.tcntl
};

/*
 * This function prints out the command and options
 */
static void print_cmd_opt(struct netdissect_options *ndo, u_char cmd, uint16_t options)
{
    char buf [100], *p = buf;

    safeputchar(ndo, cmd);

    if (options == 0)
        return;

    *p = '\0';

    if (options & XH_BTAG)
        p = strcpy (p, "Bt");
    if (options & XH_END)
        p = strcpy (p, "En");
    if (options & XH_EOM)
        p = strcpy (p, "Eo");
    if (options & XH_WCLOSE)
        p = strcpy (p, "Wc");
    if (options & XH_RCLOSE)
        p = strcpy (p, "Rc");
    if (options & XH_DREQ)
        p = strcpy (p, "Dr");
    if (options & XH_SREQ)
        p = strcpy (p, "Sr");
    if (options & XH_FASTNAK)
        p = strcpy (p, "Fn");
    if (options & XH_NOFLOW)
        p = strcpy (p, "Nf");
    if (options & XH_SORT)
        p = strcpy (p, "So");
    if (options & XH_RES)
        p = strcpy (p, "Re");
    if (options & XH_MULTI)
        p = strcpy (p, "Mc");
    if (options & XH_NOERR)
        p = strcpy (p, "Ne");
    if (options & XH_EDGE)
        p = strcpy (p, "Eg");
    if (options & XH_NOCHECK)
        p = strcpy (p, "Nc");

    ND_PRINT((ndo, "(%s)", buf));
}

void
xtp_print(netdissect_options *ndo, const u_char *bp, u_int length, const u_char *bp2)
{
    const struct xtphdr_long *xp = (struct xtphdr_long *)bp;
    const struct ip *ip = (struct ip *)bp2;
    uint16_t sport, dport;
    uint32_t sync, echo = 0;
    uint16_t options;
    xtp_seq rseq = 0L, alloc = 0;
    xtp_seq seq;
    xtp_key key, xkey;

    /*
     * We have to be careful with these next two tests.
     * xtphdr_long is defined as xtphdr plus the union of
     * xtp segment structures.  The length of xtphdr_long is
     * the length of xtphdr (32) plus the length of the largest
     * structure in the union (40).
     * We can receive packets that are packets that
     * are shorter than the xtphdr_long, e.g., the CNTL packet
     * is 64 bytes long < sizeof(xtphdr_long).
     *
     * The next two tests come from tcp_print and were
     * designed to prevent the user from looking into memory
     * if the tcp header was shortened in some way.
     * Since the common element of xtphdr_long is xtphdr, we
     * will check to see if at least sizeof(xtphdr) is safe to
     * examine.
     *
     * We must be very careful when we access values in the union,
     * as we have NOT checked it they are really there.
     *
     * awj 6/27/96
     *
     */

    if (!ND_TTEST2(xp, sizeof(struct xtphdr))) {
        ND_PRINT ((ndo,"[|xtp]"));
        return;
    }
    if (length < sizeof(struct xtphdr)) {
        ND_PRINT ((ndo,"truncated-xtp %d", length));
        return;
    }

    /*
     * Print to/from information
     */
    if (xp->xh_cmd.ptype == X_FIRST ) {
        sport = EXTRACT_BE_16BITS(&xp->xp_addr.xa_sport);
        dport = EXTRACT_BE_16BITS(&xp->xp_addr.xa_dport);
        ND_PRINT ((ndo, "%s.%s > %s.%s.",
                   ipaddr_string(ndo,&ip->ip_src),
                   tcpport_string(ndo,sport),
                   ipaddr_string(ndo,&ip->ip_dst),
                   tcpport_string(ndo,dport)));
    }
    else {
        ND_PRINT ((ndo,"%s > %s.", ipaddr_string(ndo,&ip->ip_src),
                                   ipaddr_string(ndo,&ip->ip_dst)));
    }

    /*
     * Print the key value
     */
    key = EXTRACT_BE_64BITS(&xp->xh_key);
    (void)printf("0x%" PRIx64, CLR_HI_BIT(key));
    if (TEST_HI_BIT(key)) {
        ND_PRINT ((ndo, "(r): "));
    } else {
        ND_PRINT ((ndo, ": "));
    }

    /*
     * Quiet flag, don't continue.
     */
    if (ndo->ndo_qflag) {
         ND_PRINT ((ndo, "xtp %lu", EXTRACT_BE_32BITS(&xp->xh_dlen)));
         return;
    }

    /*
     * Get the sequence number and dlen
     */
    seq = EXTRACT_BE_64BITS(&xp->xh_seq);
    length = EXTRACT_BE_32BITS(&xp->xh_dlen);

    /*
     * If this is a control packet, get the control information
     */
    if (xp->xh_cmd.ptype & 0x01) {
          rseq = EXTRACT_BE_64BITS(&xp->xp_cntl.xc_rseq);
          alloc = EXTRACT_BE_64BITS(&xp->xp_cntl.xc_alloc);
          echo = EXTRACT_BE_32BITS(&xp->xp_cntl.xc_echo);
    }

    switch(xp->xh_cmd.ptype) {
    case X_DATA:
        options = EXTRACT_BE_16BITS(&xp->xh_cmd.options);
        print_cmd_opt (ndo, '.', options);
        ND_PRINT ((ndo," %" PRId64 ":%" PRId64 "(%d)", seq, seq+length, length));
        break;
    case X_CNTL:
        options = EXTRACT_BE_16BITS(&xp->xh_cmd.options);
        print_cmd_opt (ndo, 'C', options);
        ND_PRINT ((ndo," %" PRId64 "(%d)", seq, length));
        ND_PRINT ((ndo," rseq %" PRId64 " alloc %" PRId64 " echo %lu",
                       rseq, alloc, echo));
        break;
    case X_FIRST:
        options = EXTRACT_BE_16BITS(&xp->xh_cmd.options);
        print_cmd_opt (ndo, 'F', options);
        ND_PRINT ((ndo," 0:%d(%d)", length, length));
        break;
    case X_ECNTL:
        options = EXTRACT_BE_16BITS(&xp->xh_cmd.options);
        print_cmd_opt (ndo, 'E', options);
        ND_PRINT ((ndo," %" PRId64 "(%d)", seq, length));
        ND_PRINT ((ndo," rseq %" PRId64 " alloc %" PRId64 " echo %lu",
                       rseq, alloc, echo));
        break;
    case X_TCNTL:
        options = EXTRACT_BE_16BITS(&xp->xh_cmd.options);
        print_cmd_opt (ndo, 'T', options);
        ND_PRINT ((ndo," %" PRId64 "(%d)", seq, length));
        ND_PRINT ((ndo," rseq %" PRId64 " alloc %" PRId64 " echo %lu",
                        rseq, alloc, echo));
        if (ND_TTEST(xp)) {
            xkey = EXTRACT_BE_64BITS(&xp->xp_tcntl.xc_xkey);
            ND_PRINT ((ndo," 0x%" PRIx64, CLR_HI_BIT(xkey)));
            if (TEST_HI_BIT(xkey)) {
                ND_PRINT ((ndo,"(r)"));
            }
        }
        break;
    case X_JCNTL:
        options = EXTRACT_BE_16BITS(&xp->xh_cmd.options);
        print_cmd_opt (ndo, 'J', options);
        ND_PRINT ((ndo," %" PRId64 "(%d)", seq, length));
        ND_PRINT ((ndo," rseq %" PRIu64 " alloc %" PRId64 " echo %lu",
                       rseq, alloc, echo));
        break;
    case X_DIAG:
        options = EXTRACT_BE_16BITS(&xp->xh_cmd.options);
        print_cmd_opt (ndo, 'D', options);
        ND_PRINT ((ndo," %" PRId64 "(%d)", seq, length));
        break;
    default:
        ND_PRINT ((ndo,"Error: unknown packet type: %d\n", xp->xh_cmd.ptype));
        return;
    }

    /*
     * Print the sync value
     */
    sync = EXTRACT_BE_32BITS(&xp->xh_sync);
    ND_PRINT ((ndo," sync %lu", sync));
}

#else /* if SIZEOF_LONG_LONG != 8 */

void
xtp_print (struct netdissect_options *ndo, const u_char *bp, u_int length, const u_char *bp2)
{
    const struct ip *ip = (struct ip *)bp2;

    ND_PRINT((ndo, "%s > %s %d (xtp)",
                   ipaddr_string(&ip->ip_src),
                   ipaddr_string(&ip->ip_dst), length));
}

#endif /* SIZEOF_LONG_LONG != 8 */

